/**
 *
 * \section COPYRIGHT
 *
 * Copyright 2013-2019 Software Radio Systems Limited
 *
 * \section LICENSE
 *
 * This file is part of srsLTE.
 *
 * srsUE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * srsUE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * A copy of the GNU Affero General Public License can be found in
 * the LICENSE file in the top-level directory of this distribution
 * and at http://www.gnu.org/licenses/.
 *
 */

#include "srslte/common/ota_sms.h"

namespace srslte {

// Maybe move to bcd helpers ...
static const char bcd_num_digits[] = {'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '*', '#', 'a', 'b', 'c', '\0'};

void decode_bcd_numbers_to_string(char* output, int output_len, const uint8_t* input, int input_len)
{
  int i;
  for (i = 0; i < input_len; i++) {
    /* lower nibble */
    output_len--;
    if (output_len <= 1)
      break;
    *output++ = bcd_num_digits[input[i] & 0xf];

    /* higher nibble */
    output_len--;
    if (output_len <= 1)
      break;
    *output++ = bcd_num_digits[input[i] >> 4];
  }
  if (output_len >= 1)
    *output++ = '\0';
  return;
}

bool encode_string_to_bcd_numbers(uint8_t* address, uint8_t* ptr, int* pos)
{
  char*   destination_addr_str = (char*)address;
  uint8_t addr_length          = strlen(destination_addr_str);
  int     i                    = 0;
  while (i < addr_length) {
    ptr[*pos] |= (int)destination_addr_str[i++] - 48; // 48 is ascii offset
    ptr[(*pos)++] |= ((int)destination_addr_str[i++] - 48) << 4;
  }
  if ((addr_length % 2) != 0) { // odd number of digits?
    ptr[(*pos) - 1] |= 0xF0;
  }
  return true;
}

void decode_scts(uint8_t* scts, time_t* time)
{
  struct tm tm;
  memset(&tm, 0x00, sizeof(struct tm));

  uint8_t yr = gsm411_unbcdify(*scts++);

  if (yr <= 80)
    tm.tm_year = 100 + yr;
  else
    tm.tm_year = yr;

  tm.tm_mon    = gsm411_unbcdify(*scts++) - 1;
  tm.tm_mday   = gsm411_unbcdify(*scts++);
  tm.tm_hour   = gsm411_unbcdify(*scts++);
  tm.tm_min    = gsm411_unbcdify(*scts++);
  tm.tm_sec    = gsm411_unbcdify(*scts++);
  tm.tm_gmtoff = gsm411_unbcdify(*scts++) * 60 * 15;

  *time = mktime(&tm);
}

/* Test with samples if possible! */
bool gen_validity_period(tp_vpf_t tp_vpf, unsigned long minutes)
{
  uint8_t test_buf[20];
  memset(test_buf, 0, sizeof(test_buf));
  int cnt = 0;

  switch (tp_vpf) {
    case TP_VP_NOT_PRESENT:
      break;

    case TP_ENHANCED_FORMAT: {
      tp_vp_enhanced_vpf_t enhanced_vpf = TP_VP_ENHANCED_SEMI_OCTET; // FIXME, default value

      test_buf[cnt] |= 0x40;         // bit 7: extension bit = 1
      test_buf[cnt] |= 0x20;         // bit 6: single shot sm = 1
      test_buf[cnt] |= enhanced_vpf; // bit 3-1: tp-vpf
      cnt++;

      switch (enhanced_vpf) {
        case TP_VP_ENHANCED_NOT_SPEC:
          break;

        case TP_VP_ENHANCED_RELATIVE_FORMAT: {
          if (minutes / 60 <= 12) {
            test_buf[cnt++] = (minutes / 5) - 1;
          } else if (minutes / 60 <= 24) {
            test_buf[cnt++] = ((minutes - (12 * 60)) / 30) + 143;
          } else if (minutes / (24 * 60) <= 7) {
            test_buf[cnt++] = (minutes / (24 * 60)) + 166;
          } else {
            test_buf[cnt++] = (minutes / (24 * 60 * 7)) + 192;
          }
          break;
        }

        case TP_VP_ENHANCED_RELATIVE_INT: {
          // 9.1.2.1 integer rep.
          if (minutes * 60 > 255)
            return false;
          if (minutes == 0)
            return false;
          test_buf[cnt++] = (uint8_t)minutes * 60;
          break;
        }

        case TP_VP_ENHANCED_SEMI_OCTET: {
          // 9.1.2.3 semi octet rep.
          test_buf[cnt++] = gsm411_bcdify((uint8_t)(minutes / 60)); // hours
          test_buf[cnt++] = gsm411_bcdify((uint8_t)(minutes % 60)); // minutes
          test_buf[cnt++] =
              gsm411_bcdify(0); // seconds, but doesnt make sense if the minutes variable is an unsigned long
          break;
        }
      }
      break;
    }

    case TP_RELATIVE_FORMAT: {
      if (minutes / 60 <= 12) {
        test_buf[cnt++] = (minutes / 5) - 1;
      } else if (minutes / 60 <= 24) {
        test_buf[cnt++] = ((minutes - (12 * 60)) / 30) + 143;
      } else if (minutes / (24 * 60) <= 7) {
        test_buf[cnt++] = (minutes / (24 * 60)) + 166;
      } else {
        test_buf[cnt++] = (minutes / (24 * 60 * 7)) + 192;
      }
      break;
    }

    case TP_ABSOLUTE_FORMAT: {
      time_t tstamp = (time_t)minutes * 60;
      gsm340_gen_scts(&test_buf[cnt], tstamp);
      cnt += 7;
      break;
    }
  }

  for (int i = 0; i < 20; i++) {
    printf("%02X ", test_buf[i]);
  }
  return true;
}

bool decode_dtap_data(uint8_t* ie_ptr, dtap_data_t* dtap_data)
{
  dtap_data->ti_flag                = (bool)(((*ie_ptr) & 0x80) >> 7);
  dtap_data->tio                    = (uint8_t)(((*ie_ptr) & 0x70) >> 4);
  dtap_data->protocol_discriminator = (dtap_pd_t)((*ie_ptr) & 0x0f);
  ie_ptr++;
  dtap_data->message_type = (dtap_sms_message_type) * (ie_ptr++);
  ie_ptr++; // skip user data length
  if (dtap_data->message_type != srslte::DTAP_SMS_MESSAGE_CP_DATA) {
    printf("Not CP-DATA to ms-to-network %hhx\n", dtap_data->message_type);
    return true;
  }
  return decode_rp_data(ie_ptr, &dtap_data->rp_data);
}

bool decode_rp_data(uint8_t* ie_ptr, rp_data_t* rp_data)
{
  uint8_t o_addr_length, d_addr_length;

  rp_data->message_type = (gsm_rp_message_type_t)(*(ie_ptr++));
  rp_data->reference    = *(ie_ptr++);
  if (rp_data->message_type != srslte::RP_ACK_MS_TO_NETWORK && rp_data->message_type != srslte::RP_ERROR_MS_TO_NETWORK) {
    printf("Not RP-ACK to ms-to-network %hhx\n", rp_data->message_type);
    // RP-Originator Address
    o_addr_length = *(ie_ptr++);
    if (o_addr_length > 0) {
      uint8_t o_addr_meta = *ie_ptr;
      decode_bcd_numbers_to_string(
          (char*)rp_data->originator_address, sizeof(rp_data->originator_address), ie_ptr + 1, o_addr_length - 1);
    }
    ie_ptr += o_addr_length;

    // RP-Destination Address
    d_addr_length = *(ie_ptr++);
    if (d_addr_length > 0) {
      uint8_t d_addr_meta = *ie_ptr;
      decode_bcd_numbers_to_string(
          (char*)rp_data->destination_address, sizeof(rp_data->destination_address), ie_ptr + 1, d_addr_length - 1);
    }
    ie_ptr += d_addr_length;
    ie_ptr++; // skip rp user data length
    if (rp_data->message_type == RP_DATA_MS_TO_NETWORK) {
      decode_gsm340_tpdu_to_sms(ie_ptr, &rp_data->gsm_sms, RP_DATA_MS_TO_NETWORK);
    } else {
      decode_gsm340_tpdu_to_sms(ie_ptr, &rp_data->gsm_sms, RP_DATA_NETWORK_TO_MS);
    }
  }
  return true;
}

/* decodes SMS-DELIVER & SMS-SUBMIT messages! */
bool decode_gsm340_tpdu_to_sms(uint8_t* tpdu, gsm_sms_t* sms, gsm_rp_message_type_t msg_type)
{
  uint8_t*     ie_ptr = tpdu;
  uint8_t      sms_mti, sms_mms;
  uint8_t      address_lv[SMS_ADDRESS_MAX];
  uint32_t     oa_len_bytes;
  unsigned int sms_alphabet;

  sms->mti = *ie_ptr & 0x03;

  if (msg_type == RP_DATA_MS_TO_NETWORK) {
    sms->tp_rd             = (*ie_ptr & 0x04) >> 2;
    sms->tp_vpf            = (tp_vpf_t)((*ie_ptr & 0x18) >> 3);
    sms->reply_path        = (*ie_ptr & 0x20) >> 5;
    sms->ud_hdr_ind        = (*ie_ptr & 0x40) >> 6;
    sms->tp_status_rep_req = (*ie_ptr & 0x80) >> 7;
    ie_ptr++;
    sms->msg_ref = *(ie_ptr)++;
  } else if (msg_type == RP_DATA_NETWORK_TO_MS) {
    sms->mms             = (*ie_ptr & 0x04) >> 2;
    sms->loop_prevention = (*ie_ptr & 0x08) >> 3;
    sms->reply_path      = (*ie_ptr & 0x10) >> 4;
    sms->ud_hdr_ind      = (*ie_ptr & 0x20) >> 5;
    sms->status_rep_ind  = (*ie_ptr & 0x40) >> 6;
    ie_ptr++;
  } else {
    return false;
  }

  /* length in bytes of the originate address */
  oa_len_bytes = 2 + (*ie_ptr) / 2 + (*ie_ptr) % 2;
  if (oa_len_bytes > SMS_ADDRESS_MAX) {
    printf("Originate Address > %d bytes\n", SMS_ADDRESS_MAX);
    return false;
  }

  /* convert to real number */
  if (((ie_ptr[1] & 0x70) >> 4) == 1)
    strcpy(sms->address, "+");
  else if (((ie_ptr[1] & 0x70) >> 4) == 2)
    strcpy(sms->address, "0");
  else
    sms->address[0] = '\0';

  memset(address_lv, 0, sizeof(address_lv));
  memcpy(address_lv, ie_ptr + 2, oa_len_bytes);

  decode_bcd_numbers_to_string(
      sms->address + strlen(sms->address), sizeof(sms->address) - strlen(sms->address), address_lv, oa_len_bytes - 2);

  ie_ptr += oa_len_bytes;

  sms->protocol_id        = *ie_ptr++;
  sms->data_coding_scheme = *ie_ptr++;

  sms_alphabet = gsm338_get_sms_alphabet(sms->data_coding_scheme);
  if (sms_alphabet == 0xffffffff) {
    return false;
  }

  // Only downlink contain a timestamp
  if (msg_type == RP_DATA_NETWORK_TO_MS) {
    /* get timestamp */
    decode_scts(ie_ptr, &sms->time);
    ie_ptr += 7;
  }
  /* validity period is not fully tested yet! */
  else if (msg_type == RP_DATA_MS_TO_NETWORK && sms->tp_vpf != 0x00) {
    sms->validity_minutes = gsm340_validity_period(sms->tp_vpf, ie_ptr);
    ie_ptr++;
  }

  /* user data */
  uint32_t user_data_len = *ie_ptr++;
  if (user_data_len > 0) {
    memcpy(sms->user_data, ie_ptr, user_data_len);
    switch (sms_alphabet) {
      case DCS_7BIT_DEFAULT:
        memset(sms->text, '\0', SMS_TEXT_SIZE);
        gsm_7bit_decode_n(sms->text, SMS_TEXT_SIZE, ie_ptr, user_data_len);
        break;
      case DCS_8BIT_DATA:
      case DCS_UCS2:
      case DCS_NONE:
        break;
    }
  }
  return true;
}

uint8_t gsm340_gen_addr(uint8_t* ptr, uint8_t type, uint8_t plan, char* address, int* cnt)
{

  //[FIXME] what if number starts with 00?
  if (address[0] == '+') {
    address++;
  }
  uint8_t addr_length = strlen(address);

  /* Generates an address field in line with GSM 03.40 */
  ptr[*cnt] = 0;
  ptr[*cnt] |= 0x80;      // ext = 1, bit 8
  ptr[*cnt] |= type << 4; // type of number, bit 7-5
  ptr[*cnt] |= plan;      // numbering plan identification, bit 4-1
  (*cnt)++;

  uint8_t i = 0;
  while (i < addr_length) {
    ptr[*cnt] = (uint8_t)address[i++] - 48; // 48 is ascii offset
    ptr[(*cnt)++] |= ((uint8_t)address[i++] - 48) << 4;
  }
  if ((addr_length % 2) != 0) { // odd number of digits?
    ptr[(*cnt) - 1] |= 0xF0;
  }

  return addr_length;
}

/* Encodes SMS DELIVER messages as described in 3GPP TS 23.040 V11.5.0 (2013-03) */
bool encode_tpud_sms_deliver(gsm_sms_t* sms, uint8_t* ptr, uint32_t* written_bytes, int sms_testcase)
{
  int cnt  = 0;
  ptr[cnt] = 0;
  ptr[cnt] |= sms->mti;                  // TP MTI = 00
  ptr[cnt] |= sms->mms << 2;             // TP-MMS = 1
  ptr[cnt] |= sms->loop_prevention << 3; // TP-LP = 0
  ptr[cnt] |= sms->reply_path << 7;      // TP-RP
  ptr[cnt] |= sms->ud_hdr_ind << 6;      // TP-UDHI
  ptr[cnt] |= sms->status_rep_ind << 5;  // TP-SRR
  cnt++;

  int pos_len = cnt;
  cnt++;
  /* CHECKME: type-of-number && numbering-plan TS 23.040 - 9.1.2.5 Address fields */
  // type-of-number = numbering-plan = 0 -> Unknown number, if there is no +49 at the beginning
  uint8_t addr_len = gsm340_gen_addr(ptr, 0x01, 0x01, sms->address, &cnt);
  ptr[pos_len]     = addr_len;

  ptr[cnt++] = sms->protocol_id;        // 0x00
  ptr[cnt++] = sms->data_coding_scheme; // 0x00

  /* gsm340_gen_scts generates time stamp*/
  if (sms_testcase == 14 || sms_testcase == 15) {
    for (int i = cnt; i < cnt+7; ++i) {
      ptr[i] = 0x00;
      if (i == cnt+5 && sms_testcase == 14){ 
        ptr[i] = 0x0a;
      };
    }
    cnt += 7;
  } else {
    gsm340_gen_scts(&ptr[cnt], sms->time);
    cnt += 7;
  };

  /* CHECKME: Working text length? */
  cnt++; // skip text len, set it later

  int bytes_written = 0;

  // Udh
  int udh_total_len =
      sms->user_data[0] > 0 ? sms->user_data[0] + 1 : 0; // Udh total length is udh length +1 for length indicator field
  memcpy(&ptr[cnt], sms->user_data, udh_total_len);
  // gsm_7bit_encode_n(&ptr[cnt], udh_total_len, (char *)sms->user_data, &udh_total_len);

  if (strlen(sms->text) > 0 && sms->data_coding_scheme == 0x00) {
    uint8_t* text_buffer = (uint8_t*)calloc(strlen(sms->text) * 2, sizeof(uint8_t));
    int      num_octets = gsm_septet_encode(text_buffer, sms->text);
    gsm_septet_pack(&ptr[cnt + udh_total_len], text_buffer, num_octets, udh_total_len > 0 ? (7 - ((udh_total_len * 8) % 7)) : 0);
    bytes_written = num_octets;

    // gsm_7bit_encode_n(&ptr[cnt + udh_total_len], 128, sms->text, &bytes_written); // FIXME set max of octets to be
    // written
  } else if (strlen(sms->text) > 0 && sms->data_coding_scheme == 0x04) { // GSM 8BIT
    memcpy(&ptr[cnt + udh_total_len], sms->text, strlen(sms->text));
    bytes_written = strlen(sms->text);
  } else if (strlen(sms->text) > 0 && sms->data_coding_scheme == 0x08) { // UCS2
    std::wstring_convert<std::codecvt_utf8<char16_t>, char16_t> ucs2conv;
    std::u16string ucs2 = ucs2conv.from_bytes(sms->text);
    ucs2.copy((char16_t *)&ptr[cnt + udh_total_len], ucs2.length());
    for (uint32_t i = 0; i < ucs2.length(); i++) {
      ((char16_t *)&ptr[cnt + udh_total_len])[i] = osmo_htons(((char16_t *)&ptr[cnt + udh_total_len])[i]);
    }
    bytes_written = ucs2.length() * 2;
  }

  bytes_written += udh_total_len;
  ptr[cnt - 1] = (sms->_user_data_length_override > 0) ? sms->_user_data_length_override : ((uint8_t)bytes_written); // set user data length field (includes 1 byte for the length itself)
  *written_bytes += cnt + bytes_written;
  return true;
}

/* Encodes SMS SUBMIT messages as described in 3GPP TS 23.040 V11.5.0 (2013-03) */
bool encode_tpud_sms_submit(gsm_sms_t* sms, uint8_t* ptr, uint32_t* written_bytes)
{
  int cnt  = 0;
  ptr[cnt] = 0;
  ptr[cnt] |= sms->mti;                    // TP MTI 2b
  ptr[cnt] |= sms->tp_rd << 2;             // TP-RD 1b
  ptr[cnt] |= sms->tp_vpf << 3;            // TP-LP 2b
  ptr[cnt] |= sms->reply_path << 5;        // TP-RP
  ptr[cnt] |= sms->ud_hdr_ind << 6;        // TP-UDHI
  ptr[cnt] |= sms->tp_status_rep_req << 7; // TP-SRR
  cnt++;

  ptr[cnt++] = sms->msg_ref;

  int pos_len = cnt;
  cnt++;
  uint8_t addr_len = gsm340_gen_addr(ptr, 0x01, 0x01, sms->address, &cnt);
  ptr[pos_len]     = addr_len;

  ptr[cnt++] = sms->protocol_id;
  ptr[cnt++] = sms->data_coding_scheme;

  if (sms->tp_vpf != 0) {
    gen_validity_period(sms->tp_vpf, sms->validity_minutes);
    cnt += 7;
  }

  ptr[cnt++] = (uint8_t)strlen(sms->text);

  int bytes_written = 0;
  if (strlen(sms->text) > 0) {
    /* [FIXME] Set max of octets to be written */
    gsm_7bit_encode_n(&ptr[cnt], 160, sms->text, &bytes_written);
  }

  *written_bytes += cnt + bytes_written;
  return true;
}

/* Not tested yet! */
bool encode_tpud_sms_submit_report(gsm_sms_t* sms, uint8_t* ptr, uint32_t* written_bytes)
{
  int cnt  = 0;
  ptr[cnt] = 0;
  ptr[cnt] |= sms->mti;             // TP MTI 2b
  ptr[cnt] |= sms->ud_hdr_ind << 2; // TP-UDHI
  cnt++;

  ptr[cnt++] = sms->tp_pi;

  gsm340_gen_scts(&ptr[cnt], sms->time);
  cnt += 7;

  /* take care of optional parameters */
  if (sms->tp_pi & 0x01)
    ptr[cnt++] = sms->protocol_id;
  if (sms->tp_pi & 0x02)
    ptr[cnt++] = sms->data_coding_scheme;
  if (sms->tp_pi & 0x04) {
    ptr[cnt++] = sms->tp_udl;
    if (sms->tp_udl > 0) {
      for (int i = 0; i < sms->tp_udl; i++) {
        ptr[cnt++] = sms->tp_ud[i];
      }
    }
  }
  return false;
}

/* Not tested yet! */
bool encode_tpud_sms_deliver_report(gsm_sms_t* sms, uint8_t* ptr, uint32_t* written_bytes)
{
  int cnt  = 0;
  ptr[cnt] = 0;
  ptr[cnt] |= sms->mti;
  ptr[cnt] |= sms->ud_hdr_ind << 2;
  cnt++;

  // Check 9.2.2.1a - TS 23.040 (Integer representation)
  if (sms->tp_fcs > 0) { // FIXME: does it work like that? Integer representation
    ptr[cnt++] = sms->tp_fcs;
  }

  ptr[cnt++] = sms->tp_pi;

  /* take care of optional parameters concerning the TP-Parameter-Indicator */
  if (sms->tp_pi & 0x01)
    ptr[cnt++] = sms->protocol_id;
  if (sms->tp_pi & 0x02)
    ptr[cnt++] = sms->data_coding_scheme;
  if (sms->tp_pi & 0x04) {
    ptr[cnt++] = sms->tp_udl;
    if (sms->tp_udl > 0) {
      for (int i = 0; i < sms->tp_udl; i++) {
        ptr[cnt++] = sms->tp_ud[i];
      }
    }
  }
  return true;
}

bool encode_rpdu(rp_data_t* rp_data, uint8_t* ptr, uint32_t* written_bytes, int sms_testcase)
{
  int cnt    = 0;
  ptr[cnt++] = (int8_t)rp_data->message_type;
  ptr[cnt++] = rp_data->reference;

  /* SMS-DELIVER -> originator address */
  if (rp_data->message_type == RP_DATA_NETWORK_TO_MS) {
    if (sms_testcase == 1) {
      ptr[cnt++] = 0x00; // set oridinator addres to 0x00
    } else {
      int o_addr_length = strlen((char*)rp_data->originator_address);
      if ((char)rp_data->originator_address[0] == '+') {
        o_addr_length -= 1;
      }

      if ((o_addr_length % 2) != 0) {
        ptr[cnt++] = (o_addr_length + 1 + 2) / 2; // does this work for everything?
      } else {
        ptr[cnt++] = (o_addr_length + 2) / 2; // does this work for everything?
      }
      gsm340_gen_addr(ptr, 0x01, 0x01, (char*)rp_data->originator_address, &cnt);
    }
  } else if (rp_data->message_type == RP_DATA_MS_TO_NETWORK) {
    ptr[cnt++] = 0x00; // length byte
  }

  /* SMS-SUBMIT -> destination address */
  if (rp_data->message_type == RP_DATA_MS_TO_NETWORK || sms_testcase == 4) {
    int d_addr_length = strlen((char*)rp_data->destination_address) - 1;
    if ((char)rp_data->destination_address[0] == '+') {
      d_addr_length -= 1;
    }

    /* increase cnt by number of bytes written*/
    if ((d_addr_length % 2) != 0) {
      ptr[cnt++] = (d_addr_length + 1 + 2) / 2; // does this work for everything?
    } else {
      ptr[cnt++] = (d_addr_length + 2) / 2; // does this work for everything?
    }
    gsm340_gen_addr(ptr, 0x01, 0x01, (char*)rp_data->destination_address, &cnt);
  } else if (rp_data->message_type == RP_DATA_NETWORK_TO_MS) {
    ptr[cnt++] = 0x00; // length byte
  }

  switch (rp_data->message_type) {
    case RP_DATA_MS_TO_NETWORK: {
      cnt++; // skip setting rp-user data len
      *written_bytes += cnt;
      uint8_t writtenSoFar = *written_bytes;
      bool    result       = encode_tpud_sms_submit(&rp_data->gsm_sms, &ptr[cnt], written_bytes);
      ptr[cnt - 1]         = *written_bytes - writtenSoFar; // set rp-user data len
      return result;
    }
    case RP_DATA_NETWORK_TO_MS: {
      cnt++; // skip setting rp-user data len
      *written_bytes += cnt;
      uint8_t writtenSoFar = *written_bytes;
      bool    result       = encode_tpud_sms_deliver(&rp_data->gsm_sms, &ptr[cnt], written_bytes, sms_testcase);
      if (rp_data->gsm_sms._rp_user_data_length_override != 0) {
        ptr[cnt - 1] = rp_data->gsm_sms._rp_user_data_length_override;
      } else {
        ptr[cnt - 1] = *written_bytes - writtenSoFar; // set rp-user data len
      }
      return result;
    }
    /* Not tested yet! */
    // How to determine if submit report is needed? Where is the report flag located?
    case RP_ACK_MS_TO_NETWORK: {
      ptr[cnt++] = 0x41; // RP-ACK IEI
      cnt++;             // skip setting rp-user data len, set it later!
      *written_bytes += cnt;
      uint8_t writtenSoFar = *written_bytes;
      // bool    result       = encode_tpud_sms_deliver_report(&rp_data->gsm_sms, &ptr[cnt], written_bytes);
      ptr[cnt - 1] = *written_bytes - writtenSoFar; // rp-user data len
      // return result;
      return true;
    }
    /* Not tested yet! */
    // Where is the deliver report flag located?
    case RP_ACK_NETWORK_TO_MS: {
      ptr[cnt++] = 0x41; // RP-ACK IEI
      cnt++;             // skip setting rp-user data len, set it later!
      *written_bytes += cnt;
      uint8_t writtenSoFar = *written_bytes;
      // bool    result       = encode_tpud_sms_submit_report(&rp_data->gsm_sms, &ptr[cnt], written_bytes);
      ptr[cnt - 1] = *written_bytes - writtenSoFar; // rp-user data len
      // return result;
      return true;
    }
    /* Not tested yet! */
    // Where is the deliver report flag located?
    case RP_ERROR_MS_TO_NETWORK: {
      ptr[cnt++] = 0x41; // RP-ERROR IEI
      cnt++;             // skip setting rp-user data len, set it later!
      // rp cause def.: TS 24.011 - 8.2.5.4
      cnt++; // rp cause length, set it later
      ptr[cnt++] = rp_data->gsm_sms.rp_cause;
      if (rp_data->gsm_sms.rp_err_diag != 0) {
        ptr[cnt++]   = rp_data->gsm_sms.rp_err_diag;
        ptr[cnt - 2] = 0x03; // rp cause length
      } else {
        ptr[cnt - 1] = 0x02;
      }
      *written_bytes += cnt;
      uint8_t writtenSoFar = *written_bytes;
      // bool    result       = encode_tpud_sms_deliver_report(&rp_data->gsm_sms, &ptr[cnt], written_bytes);
      ptr[cnt - 1] = *written_bytes - writtenSoFar; // rp-user data len
      // return result;
      return true;
    }
    /* Not tested yet! */
    // Where is the submit report flag located?
    case RP_ERROR_NETWORK_TO_MS: {
      ptr[cnt++] = 0x41; // RP-ERROR IEI
      cnt++;             // skip setting rp-user data len, set it later!
      // rp cause def.: TS 24.011 - 8.2.5.4
      cnt++; // rp cause length, set it later
      ptr[cnt++] = rp_data->gsm_sms.rp_cause;
      if (rp_data->gsm_sms.rp_err_diag != 0) {
        ptr[cnt++]   = rp_data->gsm_sms.rp_err_diag;
        ptr[cnt - 2] = 0x03; // rp cause length
      } else {
        ptr[cnt - 1] = 0x02;
      }
      *written_bytes += cnt;
      uint8_t writtenSoFar = *written_bytes;
      // bool    result       = encode_tpud_sms_submit_report(&rp_data->gsm_sms, &ptr[cnt], written_bytes);
      ptr[cnt - 1] = *written_bytes - writtenSoFar; // rp-user data len
      // return result;
      return true;
    }
    case RP_SMMA_MS_TO_NETWORK:
      // RP SMMA needs MTI and Message Reference
      return true;
    case RP_N_MESSAGE_TYPE_N_ITEMS:
      printf("RP ACK (MS->N) not implemented.");
      break;
  }
  return false;
}

bool encode_cp_data(dtap_data_t* dtap_data, uint8_t* ptr, uint32_t* written_bytes, int sms_testcase)
{
  int cnt  = 0;
  ptr[cnt] = 0;
  ptr[cnt] |= (dtap_data->ti_flag == true ? 0x80 : 0x00);
  ptr[cnt] |= (((dtap_data->tio & 0x07) << 4));
  ptr[cnt] |= (uint8_t)((dtap_data->protocol_discriminator) & 0x0F);
  cnt++;
  ptr[cnt++] = (uint8_t)dtap_data->message_type;

  switch (dtap_data->message_type) {
    case DTAP_SMS_MESSAGE_CP_DATA: {
      cnt++; // skip setting cp-user data len
      *written_bytes = cnt;
      bool result;
      result = encode_rpdu(&dtap_data->rp_data, &ptr[cnt], written_bytes, sms_testcase);
      ptr[2] = *written_bytes - 3; // set cp-user data len
      return result;
    }

    case DTAP_SMS_MESSAGE_CP_ERROR: {
      ptr[cnt++]     = dtap_data->cp_cause;
      *written_bytes = cnt;
      return true;
    }

    case DTAP_SMS_MESSAGE_CP_ACK: {
      *written_bytes = cnt;
      return true;
    }
  }
  return false;
}

} // namespace srslte
